"""Init for AIList."""

from __future__ import absolute_import
from .ailist import AIList, Interval
from .ailist import get_include

# This is extracted automatically by the top-level setup.py.
__version__ = '0.1.6'